@@ -3,7 +3,9 @@ package Protocol::UWSGI;
 use strict;
 use warnings;
 
-our $VERSION = '0.002';
+use parent qw(Exporter);
+
+our $VERSION = '1.000';
 
 =head1 NAME
 
@@ -11,23 +13,22 @@ Protocol::UWSGI - handle the UWSGI wire protocol
 
 =head1 VERSION
 
-Version 0.002
+Version 1.000
 
 =head1 SYNOPSIS
 
  use strict;
  use warnings;
- use Protocol::UWSGI;
- my $uwsgi = Protocol::UWSGI->new;
+ use Protocol::UWSGI qw(:all);
  # Encode...
- my $req = $uwsgi->build_request(
+ my $req = build_request(
    uri    => 'http://localhost',
    method => 'GET',
    remote => '1.2.3.4:1234',
  );
  # ... and decode again
- warn "URI was " . $uwsgi->uri_from_env(
-   $uwsgi->extract_frame(\$req)
+ warn "URI was " . uri_from_env(
+   extract_frame(\$req)
  );
 
 =head1 DESCRIPTION
@@ -44,14 +45,14 @@ all. Try L<Net::Async::UWSGI> for an implementation that actually does
 something useful.
 
 Typically you'd create a UNIX socket and listen for requests, passing
-any data to the L</extract_frame> method and handling the resulting
-data if that method returns something other than undef:
+any data to the L</extract_frame> function and handling the resulting
+data if that function returns something other than undef:
 
  # Detect read - first packet is usually the UWSGI header, everything
  # after that would be the HTTP request body if there is one:
  sub on_read {
    my ($self, $buffref) = @_;
-   while(my $pkt = $uwsgi->extract_frame($buffref)) {
+   while(my $pkt = extract_frame($buffref)) {
      $self->handle_uwsgi($pkt);
    }
  }
@@ -59,9 +60,28 @@ data if that method returns something other than undef:
  # and probably an EOF handler to detect client hangup
  # sub on_eof { ... }
 
+=head1 IMPLEMENTATION - Server
+
+A server implementation typically accepts requests from a reverse
+proxy, such as nginx, and returns HTTP responses.
+
+Import the :server tag to get L</uri_from_env>, L</extract_frame>
+and in future maybe L</psgi_from_env> functions:
+
+ use Protocol::UWSGI qw(:server);
+
+=head1 IMPLEMENTATION - Client
+
+A client implementation typically accepts HTTP requests and converts
+them to UWSGI for passing to a UWSGI-capable application.
+
+Import the :client tag to get L</build_request>:
+
+ use Protocol::UWSGI qw(:client);
+
 =cut
 
-use Encode;
+use Encode ();
 use URI;
 
 use constant {
@@ -69,7 +89,22 @@ use constant {
 	PSGI_MODIFIER2 => 0,
 };
 
-=head1 METHODS
+our @EXPORT_OK = qw(
+	extract_frame
+	uri_from_env
+
+	build_request
+
+	PSGI_MODIFIER1
+	PSGI_MODIFIER2
+);
+our %EXPORT_TAGS = (
+	'server' => [qw(extract_frame uri_from_env)],
+	'client' => [qw(build_request)],
+	'all' => \@EXPORT_OK
+);
+
+=head1 FUNCTIONS
 
 If you're handling incoming UWSGI requests, you'll need to instantiate
 via L</new> then decode the request using L</extract_frame>.
@@ -81,14 +116,6 @@ Just want to decode captured traffic? L</extract_frame> again.
 
 =cut
 
-=head2 new
-
-Instantiate. Takes no parameters, returns an instance.
-
-=cut
-
-sub new { my $class = shift; bless { @_ }, $class }
-
 =head2 extract_frame
 
 Attempts to extract a single UWSGI packet from the given buffer (which
@@ -108,10 +135,32 @@ the buffer is guaranteed not to be modified.
 This may be called as a class method or an instance method.
 The instance state will remain unchanged after calling this method.
 
+Note that there is no constructor provided in this
+class - if you want to call this as an instance method,
+you'll need to bless manually or be applying this as
+a role/mixin.
+
 =cut
 
 sub extract_frame {
-	my $self = shift;
+	my ($buffref) = @_;
+
+	my ($modifier1, $length, $modifier2) = unpack 'C1v1C1', $$buffref;
+	# no, still too short
+	return undef unless $length && length $$buffref >= $length + 4;
+
+	# then do the modifier-specific handling
+	die "Unsupported modifier1 $modifier1" unless $modifier1 == PSGI_MODIFIER1;
+
+	# hack bits off the buffer
+	substr $$buffref, 0, 4, '';
+
+	my %env = unpack '(v1/a*)*', substr $$buffref, 0, $length, '';
+	\%env
+}
+
+# For cases where non-PSGI modifiers are wanted. Takes about 2.5x as long.
+sub extract_frame_universal {
 	my $buffref = shift;
 	# too short
 	return undef unless length $$buffref >= 4;
@@ -124,7 +173,7 @@ sub extract_frame {
 	substr $$buffref, 0, 4, '';
 
 	# then do the modifier-specific handling
-	return $self->extract_modifier(
+	return extract_modifier(
 		modifier1 => $modifier1,
 		modifier2 => $modifier2,
 		length    => $length,
@@ -141,7 +190,6 @@ If zero, this means we should be able to extract a valid frame.
 =cut
 
 sub bytes_required {
-	my $self = shift;
 	my $buffref = shift;
 	return 4 - length($$buffref) unless length $$buffref >= 4;
 
@@ -178,7 +226,6 @@ Returns a scalar containing packet data or raises an exception on failure.
 =cut
 
 sub build_request {
-	my $self = shift;
 	my %args = @_;
 
 #	my $type = delete $args{type} or die 'no type provided';
@@ -186,30 +233,33 @@ sub build_request {
 	$uri = URI->new($uri) unless ref $uri;
 
 	my %env;
-	$env{REQUEST_METHOD} = uc $args{method};
+	$env{REQUEST_METHOD} = uc delete $args{method};
 	$env{UWSGI_SCHEME} = $uri->scheme;
 	$env{HTTP_HOST} = $uri->host;
 	$env{SERVER_PORT} = $uri->port // 80;
 	$env{PATH_INFO} = $uri->path;
 	$env{QUERY_STRING} = $uri->query if defined $uri->query;
-	@env{qw(REMOTE_ADDR REMOTE_PORT)} = split ':', $args{remote}, 2;
+	@env{qw(REMOTE_ADDR REMOTE_PORT)} = split ':', delete $args{remote}, 2 if $args{remote};
 
 	$args{headers} ||= {};
 	foreach my $k (keys %{$args{headers}}) {
 		(my $env_k = uc $k) =~ tr/-/_/;
 		$env{"HTTP_$env_k"} = $args{headers}{$k} // '';
 	}
+	delete $args{headers};
 
+	my @modifier = delete @args{qw(modifier1 modifier2)};
 	my $data = '';
+	%env = (%args, %env);
 	foreach my $k (sort keys %env) {
 		die "Undef value found for $k" unless defined $env{$k};
 		$data .= pack 'v1/av1/a', map { Encode::encode('utf8', $_) } $k, $env{$k};
 	}
 
 	return pack('C1v1C1',
-		$args{modifier1} // PSGI_MODIFIER1,
+		$modifier[0] // PSGI_MODIFIER1,
 		length($data),
-		$args{modifier2} // PSGI_MODIFIER2,
+		$modifier[1] // PSGI_MODIFIER2,
 	) . $data;
 }
 
@@ -220,7 +270,6 @@ Used internally to extract and handle the modifier-specific data.
 =cut
 
 sub extract_modifier {
-	my $self = shift;
 	my %args = @_;
 
 	die "Unsupported modifier1 $args{modifier1}" unless $args{modifier1} == PSGI_MODIFIER1;
@@ -245,7 +294,7 @@ Returns a L<URI> object parsed from a request ("environment").
 =cut
 
 sub uri_from_env {
-	my ($self, $env) = @_;
+	my ($env) = @_;
 	my $uri = $env->{UWSGI_SCHEME} . '://' . $env->{HTTP_HOST} . ':' . $env->{SERVER_PORT} . $env->{PATH_INFO};
 	$uri .= '?' . $env->{QUERY_STRING} if length($env->{QUERY_STRING} // '');
 	return URI->new($uri);
